#!/usr/bin/python

# Copyright (c) 2004 Joseph Walton <joe@kafsemo.org>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# $Id: unpolluted.py,v 1.11 2004/09/12 00:11:10 joe Exp $

# Check an HTTP resource for conformance to best-practices for
#  pollable resources (specifically, RSS feeds).
#
# Makes sure that necessary metadata (Date, Last-Modified, etc.) is provided,
#  that compression is handled, and that conditional requests work.
# If you're expecting a large number of clients, passing all these tests
#  is a Good Idea.


from sys import argv, stderr, exit
import urllib2
from re import search

import rfc822
import calendar
import time
import sha

def parseDate(d):
  t = rfc822.parsedate(d)
  if t:
    return calendar.timegm(t)
  else:
    return None

# Hash the resource stream, to identify changes
def contentHash(s):
  h = sha.sha()
  contentLength = 0

  d = s.read(16384)
  while d:
    contentLength += len(d)
    h.update(d)
    d = s.read(16384)

  return (h.hexdigest(), contentLength)


if len(argv) != 2:
  print >>stderr,'Usage: unpolluted <HTTP url>'
  exit(5)

url = argv[1]

def check(t,desc):
  if t:
    l = '[pass] ' + desc
  elif t == False:
    l = '[FAIL] ' + desc
  else:
    l = '[    ] ' + desc
  print l
  return t

def warnUnless(t,desc):
  if t:
    print '[good] ' + desc
  else:
    print '[WARN] ' + desc

# Indicate a mismatch with the HTTP spec
def showError(desc):
  print '[ERR ] ' + desc

def getDateHeader(s,name):
  hs = s.headers.get(name, None)
  if hs:
    t = parseDate(hs)
    if t:
      return t
    else:
      showError('HTTP dates must be formatted correctly')
  else:
    return None

print 'URL:',url

x = search(r'\d+(\.\d+)*', '$Revision: 1.11 $')
if x:
  version = x.group(0)
else:
  version = '0.0'

# Make a basic request
def makeRequest():
  req = urllib2.Request(url)
  req.add_header('Accept-Encoding', 'gzip, deflate')
  req.add_header('User-Agent', 'unpolluted/' + version + ' urllib2/' + urllib2.__version__)
  return req


# A simple request, with no prior state

s = urllib2.urlopen(makeRequest())


# Should verify status == 200; presume exceptions will be raised otherwise


# Warn about redirection
check(url == s.geturl(), "The URL shouldn't redirect")


# Make sure compression is used
comp = s.headers.get('Content-Encoding', None)
check(None != comp, 'If clients ask for compression, then provide it')


# Check for Cache-Control
cacheControl = s.headers.get('Cache-Control')

ccp = True

if cacheControl:
  ccl = urllib2.parse_http_list(cacheControl.lower())
  for b in ['private', 'no-cache', 'no-store']:
    ccp &= b not in ccl
  
check(ccp, "Don't disable caching with Cache-Control")


# Check for a valid Date header
date = getDateHeader(s, 'Date')

check(date != None, 'Supply a Date header, to indicate the server time')

dc = None

if date:
  dc = abs(time.time() - date) < (60 * 10)
else:
  # We will use this later, so take the correct time for comparison
  date = time.time()

check(dc, 'The server clock should be accurate (< 10 minutes difference)')


# Check for an Expires header
expires = getDateHeader(s, 'Expires')

check(expires != None, 'Use an Expires header to tell clients how often to poll')

if check(expires and (expires > date), 'Expiry must be in the future'):
  warnUnless(expires - date >= (60 * 15), 'Expiry should be at least fifteen minutes into the future')

# Check for a plausible Last-Modified
lastModified = getDateHeader(s, 'Last-Modified')

check(lastModified != None, 'Last-Modified headers tell clients how often content is updated')

check(lastModified and (lastModified < date), 'The Last-Modified date must be in the past')


# Make sure an ETag is supplied
etag = s.headers.get('ETag', None)
warnUnless(etag != None, 'Provide an ETag')


# They must have supplied an ETag *or* Last-Modified
check(lastModified or etag, 'Provide Last-Modified or ETag, so clients can make conditional requests')


# Check the content length, and store a hash
(hd, contentLength) = contentHash(s)

warnUnless(contentLength <= (64 * 1024), 'Feed size should be less than 64 KiB')



# Create an opener that accepts, and flags, 304 status codes
#  (rather than raising an exception)

class NotModifiedHandler(urllib2.BaseHandler):
  def http_error_304(self, req, fp, code, message, headers):
    addinfourl = urllib2.addinfourl(fp, headers, req.get_full_url())
    addinfourl.code = code
    return addinfourl

opener = urllib2.build_opener(NotModifiedHandler())



# If we have Last-Modified, or an ETag, supply both for a conditional request
conditionalSuccess = None

if lastModified or etag:
  req = makeRequest()
  if lastModified:
    req.add_header('If-Modified-Since', rfc822.formatdate(lastModified))
  if etag:
    req.add_header('If-None-Match', etag)
  s = opener.open(req)
  hdrs = s.info()
  conditionalSuccess = hasattr(s, 'code') and s.code == 304
  if not(conditionalSuccess) and (hd, contentLength) != contentHash(s):
    showError('Content has changed between polls; unable to verify conditional request')

check(conditionalSuccess, 'Conditional requests, using Last-Modified and ETag, save bandwidth and reduce server load')


# If we have a Last-Modified date, test support for If-Modified-Since
lastModSupported = None
if lastModified:
  req = makeRequest()
  req.add_header('If-Modified-Since', rfc822.formatdate(lastModified))

  s = opener.open(req)
  hdrs = s.info()
  lastModSupported = hasattr(s, 'code') and s.code == 304
  if not(lastModSupported):
    if (hd, contentLength) != contentHash(s):
      showError('Content has changed between polls; unable to verify If-Modified-Since support')
  warnUnless(lastModSupported, 'If-Modified-Since alone should be honoured')



# If we have an ETag, test support for If-None-Match
ifNoneMatchSupported = None
if etag:
  req = makeRequest()
  req.add_header('If-None-Match', etag)

  s = opener.open(req)
  hdrs = s.info()
  ifNoneMatchSupported = hasattr(s, 'code') and s.code == 304
  if not(ifNoneMatchSupported):
    if (hd, contentLength) != contentHash(s):
      showError('Content has changed between polls; unable to verify If-None-Match support')
  warnUnless(ifNoneMatchSupported, 'If-None-Match alone should be honoured')
